/*
 * Common clock functionality.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#ifdef __cplusplus

#include <l4/devices/hw_register_block.h>
#include <l4/devices/cpm.h>
#include <l4/sys/types.h>
#include <stdint.h>

/* Forward declaration. */

class Clock_base;



/* Register access type. */

class Cpm_regs
{
  Hw::Register_block<32> _regs;

protected:
  Clock_base **_clocks;

public:
  uint32_t exclk_freq;

  explicit Cpm_regs(l4_addr_t addr, Clock_base *clocks[],
                    uint32_t exclk_freq);

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);

  Clock_base *get_clock(int num);
};



// Register field abstraction.

class Field
{
  uint32_t reg;
  uint32_t mask;
  uint8_t bit;
  bool defined;

public:
  explicit Field()
  : defined(false)
  {
  }

  explicit Field(uint32_t reg, uint32_t mask, uint32_t bit)
  : reg(reg), mask(mask), bit(bit), defined(true)
  {
  }

  uint32_t get_field(Cpm_regs &regs);
  void set_field(Cpm_regs &regs, uint32_t value);
  bool is_defined() { return defined; }

  // Undefined field object.

  static Field undefined;
};



// Clock sources.

class Mux
{
  int _num_inputs;
  enum Clock_identifiers *_inputs, _input;

public:
  explicit Mux(int num_inputs, enum Clock_identifiers inputs[])
  : _num_inputs(num_inputs), _inputs(inputs)
  {
  }

  explicit Mux(enum Clock_identifiers input)
  : _num_inputs(1), _inputs(&_input)
  {
    _input = input;
  }

  explicit Mux()
  : _num_inputs(0), _inputs(NULL)
  {
  }

  int get_number() { return _num_inputs; }
  enum Clock_identifiers get_input(int num);
};



class Source
{
  Mux _inputs;
  Field _source;

public:
  explicit Source(Mux inputs, Field source)
  : _inputs(inputs), _source(source)
  {
  }

  explicit Source(Mux inputs)
  : _inputs(inputs)
  {
  }

  explicit Source()
  {
  }

  int get_number() { return _inputs.get_number(); }
  enum Clock_identifiers get_input(int num) { return _inputs.get_input(num); }

  // Clock source.

  uint8_t get_source(Cpm_regs &regs);
  void set_source(Cpm_regs &regs, uint8_t source);

  // Clock source frequency.

  uint32_t get_frequency(Cpm_regs &regs);

  // Undefined source object.

  static Source undefined;
};



// Frequency transformation.

class Transform
{
public:

  // Output frequency.

  virtual uint32_t get_frequency(Cpm_regs &regs, uint32_t source_frequency) = 0;
};



class Divider : public Transform
{
  Field _divider;

public:
  explicit Divider(Field divider)
  : _divider(divider)
  {
  }

  explicit Divider()
  : _divider(Field::undefined)
  {
  }

  // Clock divider.

  uint32_t get_divider(Cpm_regs &regs);
  void set_divider(Cpm_regs &regs, uint32_t division);

  // Output frequency.

  uint32_t get_frequency(Cpm_regs &regs, uint32_t source_frequency);

  // Undefined divider.

  static Divider undefined;
};



class Divider_pll : public Transform
{
  Field _multiplier, _input_division, _output_division0, _output_division1;

public:
  explicit Divider_pll(Field multiplier, Field input_division,
                       Field output_division0, Field output_division1)
  : _multiplier(multiplier), _input_division(input_division),
    _output_division0(output_division0), _output_division1(output_division1)
  {
  }

  // General frequency modifiers.

  uint16_t get_multiplier(Cpm_regs &regs);
  void set_multiplier(Cpm_regs &regs, uint16_t multiplier);
  uint8_t get_input_division(Cpm_regs &regs);
  void set_input_division(Cpm_regs &regs, uint8_t divider);
  uint8_t get_output_division(Cpm_regs &regs);
  void set_output_division(Cpm_regs &regs, uint8_t divider);

  // Output frequency.

  uint32_t get_frequency(Cpm_regs &regs, uint32_t source_frequency);

  // Other operations.

  void set_pll_parameters(Cpm_regs &regs, uint16_t multiplier,
                          uint8_t in_divider, uint8_t out_divider);
};



// Common clock abstraction.

class Clock_base
{
  Source _source;

public:
  explicit Clock_base(Source source)
  : _source(source)
  {
  }

  // Clock control.

  virtual int have_clock(Cpm_regs &regs);
  virtual void start_clock(Cpm_regs &regs);
  virtual void stop_clock(Cpm_regs &regs);

  // Clock divider.

  virtual uint32_t get_divider(Cpm_regs &regs);
  virtual void set_divider(Cpm_regs &regs, uint32_t division);

  // Clock source.

  virtual uint8_t get_source(Cpm_regs &regs);
  virtual void set_source(Cpm_regs &regs, uint8_t source);

  // Clock source frequency.

  virtual uint32_t get_source_frequency(Cpm_regs &regs);

  // Output frequency.

  virtual uint32_t get_frequency(Cpm_regs &regs);
};



// PLL descriptions.

class Pll : public Clock_base
{
  Field _enable, _stable, _bypass;
  Divider_pll _divider;

public:
  explicit Pll(Source source,
               Field enable, Field stable, Field bypass,
               Divider_pll divider)
  : Clock_base(source),
    _enable(enable), _stable(stable), _bypass(bypass),
    _divider(divider)
  {
  }

  // PLL_specific control.

  int have_pll(Cpm_regs &regs);
  int pll_enabled(Cpm_regs &regs);
  int pll_bypassed(Cpm_regs &regs);

  // Clock control.

  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // General frequency modifiers.

  uint16_t get_multiplier(Cpm_regs &regs);
  void set_multiplier(Cpm_regs &regs, uint16_t multiplier);
  uint8_t get_input_division(Cpm_regs &regs);
  void set_input_division(Cpm_regs &regs, uint8_t divider);
  uint8_t get_output_division(Cpm_regs &regs);
  void set_output_division(Cpm_regs &regs, uint8_t divider);

  // PLL output frequency.

  uint32_t get_frequency(Cpm_regs &regs);

  // Other operations.

  void set_pll_parameters(Cpm_regs &regs, uint16_t multiplier,
                          uint8_t in_divider, uint8_t out_divider);
};



// Clock descriptions.

class Clock : public Clock_base
{
  Field _gate, _change_enable, _busy;
  Divider _divider;

  // Clock control.

  void change_disable(Cpm_regs &regs);
  void change_enable(Cpm_regs &regs);
  void wait_busy(Cpm_regs &regs);

public:
  explicit Clock(Source source = Source::undefined,
                 Field gate = Field::undefined,
                 Field change_enable = Field::undefined,
                 Field busy = Field::undefined,
                 Divider divider = Divider::undefined)
  : Clock_base(source),
    _gate(gate), _change_enable(change_enable), _busy(busy), _divider(divider)
  {
  }

  // Clock control.

  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // Clock divider.

  uint32_t get_divider(Cpm_regs &regs);
  void set_divider(Cpm_regs &regs, uint32_t division);

  // Clock source.

  void set_source(Cpm_regs &regs, uint8_t source);

  // Output frequency.

  uint32_t get_frequency(Cpm_regs &regs);
};

#endif /* __cplusplus */
